/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_surf.c: surface-related refresh code

#include "quakedef.h"

GLuint r_surfacesvbo;
int r_numsurfaceverts = 0;


/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (texture_t *base)
{
	int		reletive;
	int		count;

	if (currententity->frame)
	{
		if (base->alternate_anims)
			base = base->alternate_anims;
	}

	if (!base->anim_total)
		return base;

	reletive = (int) (cl.time * 10) % base->anim_total;

	count = 0;
	while (base->anim_min > reletive || base->anim_max <= reletive)
	{
		base = base->anim_next;
		if (!base)
			Sys_Error ("R_TextureAnimation: broken cycle");
		if (++count > 100)
			Sys_Error ("R_TextureAnimation: infinite cycle");
	}

	return base;
}


/*
=============================================================

BRUSH MODELS

=============================================================
*/

#define MAX_SURFINDEXES		65536

unsigned r_surfindexes[MAX_SURFINDEXES];
int r_numsurfindexes;


void GL_BeginDrawingSurfaces (void)
{
	brushpolyvert_t *v = NULL;

	GL_EnableVertexAttribArrays (VAA0 | VAA1 | VAA2);

	glBindBuffer (GL_ARRAY_BUFFER, r_surfacesvbo);

	glVertexAttribPointer (0, 3, GL_FLOAT, GL_FALSE, sizeof (brushpolyvert_t), v->position);
	glVertexAttribPointer (1, 3, GL_FLOAT, GL_FALSE, sizeof (brushpolyvert_t), v->normal);
	glVertexAttribPointer (2, 2, GL_FLOAT, GL_FALSE, sizeof (brushpolyvert_t), v->texcoord);

	glBindBuffer (GL_ARRAY_BUFFER, 0);

	r_numsurfindexes = 0;
}


void GL_DrawSurface (msurface_t *surf)
{
	int v;
	unsigned *ndx;

	if ((r_numsurfindexes + surf->numindexes) >= MAX_SURFINDEXES)
		GL_EndDrawingSurfaces ();

	ndx = &r_surfindexes[r_numsurfindexes];

	for (v = 2; v < surf->numvertexes; v++, ndx += 3)
	{
		ndx[0] = surf->firstvertex;
		ndx[1] = surf->firstvertex + (v - 1);
		ndx[2] = surf->firstvertex + v;
	}

	r_numsurfindexes += surf->numindexes;
}


void GL_EndDrawingSurfaces (void)
{
	if (r_numsurfindexes)
	{
		glDrawElements (GL_TRIANGLES, r_numsurfindexes, GL_UNSIGNED_INT, r_surfindexes);
		r_numsurfindexes = 0;
	}
}


extrasurface_t r_extrasurfaces[MAX_EXTRA_SURFACES];
int r_numextrasurfaces;

void R_AddExtraSurface (msurface_t *surf)
{
	r_extrasurfaces[r_numextrasurfaces].surf = surf;
	r_extrasurfaces[r_numextrasurfaces].ent = currententity;

	r_numextrasurfaces++;
}


void R_DrawLumaChain (msurface_t *surf)
{
	texture_t *lasttex = NULL;

	GL_BindPrograms (r_fullbrightsurfvp, r_fullbrightfp);
	GL_BlendState (BS_MAXBLEND);
	GL_DepthState (DS_TESTEQUAL);

	for (; surf; surf = surf->texturechain)
	{
		if (surf->texinfo->texture != lasttex)
		{
			GL_EndDrawingSurfaces ();
			GL_BindTexture (1, R_TextureAnimation (surf->texinfo->texture)->fb_texturenum);
			lasttex = surf->texinfo->texture;
		}

		GL_DrawSurface (surf);
	}

	GL_EndDrawingSurfaces ();
}


void R_DrawExtraSurfaces (void)
{
	int i;

	msurface_t *skychain = NULL;
	msurface_t *waterchain = NULL;
	msurface_t *lumachain = NULL;

	if (!r_numextrasurfaces) return;

	currententity = NULL;

	GL_BeginDrawingSurfaces ();

	for (i = 0; i < r_numextrasurfaces; i++)
	{
		msurface_t *surf = r_extrasurfaces[i].surf;

		if (r_extrasurfaces[i].ent != currententity)
		{
			if (skychain)
			{
				R_DrawSkyChain (skychain);
				skychain = NULL;
			}

			if (waterchain)
			{
				R_DrawWaterChain (waterchain);
				waterchain = NULL;
			}

			if (lumachain)
			{
				R_DrawLumaChain (lumachain);
				lumachain = NULL;
			}

			// switch entity (matrix transform/etc
			glLoadMatrixf (r_world_matrix.m16);
			glMultMatrixf (r_extrasurfaces[i].ent->localmatrix.m16);

			currententity = r_extrasurfaces[i].ent;
		}

		if (surf->flags & SURF_DRAWSKY)
		{
			surf->texturechain = skychain;
			skychain = surf;
		}
		else if (surf->flags & SURF_DRAWTURB)
		{
			surf->texturechain = waterchain;
			waterchain = surf;
		}
		else
		{
			surf->texturechain = lumachain;
			lumachain = surf;
		}
	}

	if (skychain)
	{
		R_DrawSkyChain (skychain);
		skychain = NULL;
	}

	if (waterchain)
	{
		R_DrawWaterChain (waterchain);
		waterchain = NULL;
	}

	if (lumachain)
	{
		R_DrawLumaChain (lumachain);
		lumachain = NULL;
	}

	GL_EndDrawingSurfaces ();

	glLoadMatrixf (r_world_matrix.m16);
	r_numextrasurfaces = 0;
}


void R_DrawSurface_DepthPass (msurface_t *surf)
{
	// drawn separately
	if (surf->flags & (SURF_DRAWTURB | SURF_DRAWSKY))
	{
		R_AddExtraSurface (surf);
		return;
	}
	else if (R_TextureAnimation (surf->texinfo->texture)->fb_texturenum)
	{
		// fixme - animation
		R_AddExtraSurface (surf);
	}

	GL_DrawSurface (surf);
}


void R_DrawSolidChain_LightPass (msurface_t *s, texture_t *t)
{
	t = R_TextureAnimation (t);

	GL_BindTexture (0, t->gl_texturenum);

	for (; s; s = s->texturechain)
	{
		GL_DrawSurface (s);
	}

	GL_EndDrawingSurfaces ();
}


void R_DrawTextureChains_LightPass (model_t *mod, glmatrix *localmatrix)
{
	int		i;
	msurface_t	*s;
	texture_t	*t;

	glPushMatrix ();
	glMultMatrixf (localmatrix->m16);

	GL_BeginDrawingSurfaces ();

	for (i = 0; i < mod->numtextures; i++)
	{
		if ((t = mod->textures[i]) == NULL) continue;
		if ((s = t->texturechain) == NULL) continue;

		if (s->flags & SURF_DRAWSKY)
		{
		}
		else if (s->flags & SURF_DRAWTURB)
		{
		}
		else R_DrawSolidChain_LightPass (s, t);

		t->texturechain = NULL;
	}

	GL_EndDrawingSurfaces ();

	glPopMatrix ();
}


void R_ClearTextureChains (model_t *mod)
{
	int i;
	texture_t	*t;

	for (i = 0; i < mod->numtextures; i++)
	{
		if ((t = mod->textures[i]) == NULL) continue;
		t->texturechain = NULL;
	}
}


/*
=================
R_DrawBrushModel
=================
*/
void R_DrawBrushModel (entity_t *e)
{
	int			i;
	msurface_t	*surf;
	model_t		*clmodel;

	if (e->bbculled) return;

	currententity = e;

	clmodel = e->model;

	// deferred to here because r_refdef isn't known yet when the entity is added
	R_InverseTransform (&e->localmatrix, modelorg, r_refdef.vieworigin);

	surf = &clmodel->surfaces[clmodel->firstmodelsurface];

	glPushMatrix ();
	glMultMatrixf (e->localmatrix.m16);

	GL_BeginDrawingSurfaces ();
	GL_BindPrograms (r_depthpasssurfvp, r_depthpassfp);

	// draw texture
	for (i = 0; i < clmodel->nummodelsurfaces; i++, surf++)
	{
		// find which side of the node we are on
		mplane_t *pplane = surf->plane;
		float dot = DotProduct (modelorg, pplane->normal) - pplane->dist;

		// draw the polygon
		if (((surf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) || (!(surf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON)))
		{
			surf->visframe = r_framecount;
			R_DrawSurface_DepthPass (surf);
		}
	}

	GL_EndDrawingSurfaces ();

	glPopMatrix ();
}


void R_SetupBrushModel (entity_t *e)
{
	R_IdentityMatrix (&e->localmatrix);
	R_TranslateMatrix (&e->localmatrix, e->origin[0], e->origin[1], e->origin[2]);
	R_RotateMatrix (&e->localmatrix, e->angles[0], e->angles[1], e->angles[2]);
}


/*
=============================================================

WORLD MODEL

=============================================================
*/

/*
================
R_RecursiveWorldNode
================
*/
void R_RecursiveWorldNode (mnode_t *node)
{
	int			i, c, side, *pindex;
	vec3_t		acceptpt, rejectpt;
	mplane_t	*plane;
	msurface_t	*surf, **mark;
	mleaf_t		*pleaf;
	double		d, dot;
	vec3_t		mins, maxs;

	if (node->contents == CONTENTS_SOLID)
		return;		// solid

	if (node->visframe != r_visframecount)
		return;

	if (R_CullBox (node->minmaxs, node->minmaxs + 3))
		return;

	// if a leaf node, draw stuff
	if (node->contents < 0)
	{
		pleaf = (mleaf_t *) node;

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}

		// deal with model fragments in this leaf
		if (pleaf->efrags) R_StoreEfrags (&pleaf->efrags);
		if (pleaf->lfrags) R_StoreLfrags (&pleaf->lfrags);

		return;
	}

	// node is just a decision point, so go down the apropriate sides

	// find which side of the node we are on
	plane = node->plane;

	switch (plane->type)
	{
	case PLANE_X:
		dot = modelorg[0] - plane->dist;
		break;
	case PLANE_Y:
		dot = modelorg[1] - plane->dist;
		break;
	case PLANE_Z:
		dot = modelorg[2] - plane->dist;
		break;
	default:
		dot = DotProduct (modelorg, plane->normal) - plane->dist;
		break;
	}

	if (dot >= 0)
		side = 0;
	else
		side = 1;

	// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

	// draw stuff
	c = node->numsurfaces;

	if (c)
	{
		surf = node->surfaces;

		for (; c; c--, surf++)
		{
			if (surf->visframe != r_framecount)
				continue;

			if ((dot < 0) ^ !!(surf->flags & SURF_PLANEBACK))
				continue;		// wrong side

			R_DrawSurface_DepthPass (surf);
		}
	}

	// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}



/*
=============
R_DrawWorld
=============
*/
void R_DrawWorld (void)
{
	R_IdentityMatrix (&r_worldentity.localmatrix);

	VectorCopy (r_refdef.vieworigin, modelorg);

	currententity = &r_worldentity;

	r_numextrasurfaces = 0;

	GL_BeginDrawingSurfaces ();
	GL_BindPrograms (r_depthpasssurfvp, r_depthpassfp);

	R_RecursiveWorldNode (cl.worldmodel->nodes);

	GL_EndDrawingSurfaces ();
}


/*
===============
R_MarkLeaves
===============
*/
void R_MarkLeaves (void)
{
	byte	*vis;
	mnode_t	*node;
	int		i;
	byte	solid[4096];

	if (r_oldviewleaf == r_viewleaf && !r_novis.value)
		return;

	r_visframecount++;
	r_oldviewleaf = r_viewleaf;

	if (r_novis.value)
	{
		vis = solid;
		memset (solid, 0xff, (cl.worldmodel->numleafs + 7) >> 3);
	}
	else
		vis = Mod_LeafPVS (r_viewleaf, cl.worldmodel);

	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];
			do
			{
				if (node->visframe == r_visframecount)
					break;
				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}
}


/*
==================
GL_BuildSurfaces

==================
*/
void GL_BuildSurfaces (void)
{
	int	i, j;
	brushpolyvert_t *verts;

	r_framecount = 1;

	// build the VBO for this bunch of surfaces
	glDeleteBuffers (1, &r_surfacesvbo);

	// synchronize to ensure that the buffer object has completed destroying before continuing
	glFinish ();

	glGenBuffers (1, &r_surfacesvbo);
	glBindBuffer (GL_ARRAY_BUFFER, r_surfacesvbo);
	glBufferData (GL_ARRAY_BUFFER, r_numsurfaceverts * sizeof (brushpolyvert_t), NULL, GL_STATIC_DRAW);

	// synchronize to ensure that the buffer object has completed creation before continuing
	glFinish ();

	for (j = 1; j < MAX_MODELS; j++)
	{
		model_t	*mod = cl.model_precache[j];

		if (!mod) break;
		if (mod->name[0] == '*') continue;

		for (i = 0; i < mod->numsurfaces; i++)
		{
			msurface_t *surf = &mod->surfaces[i];
			int	v;
			float *vec;

			int mbr_offset = surf->firstvertex * sizeof (brushpolyvert_t);
			int mbr_size = surf->numvertexes * sizeof (brushpolyvert_t);
			int mbr_access = GL_MAP_WRITE_BIT | GL_MAP_UNSYNCHRONIZED_BIT | GL_MAP_INVALIDATE_RANGE_BIT;

			brushpolyvert_t *verts = (brushpolyvert_t *) glMapBufferRange (GL_ARRAY_BUFFER, mbr_offset, mbr_size, mbr_access);

			// reconstruct the polygon
			for (v = 0; v < surf->numvertexes; v++)
			{
				// get position
				int lindex = mod->surfedges[surf->firstedge + v];

				if (lindex > 0)
					vec = mod->vertexes[mod->edges[lindex].v[0]].position;
				else vec = mod->vertexes[mod->edges[-lindex].v[1]].position;

				Vector3Copy (verts[v].position, vec);

				// copy over normal, adjusting for planeback
				if (surf->flags & SURF_PLANEBACK)
					Vector3Scalef (verts[v].normal, surf->plane->normal, -1.0f);
				else Vector3Copy (verts[v].normal, surf->plane->normal);

				// calc texcoords
				if (surf->flags & SURF_DRAWTURB)
				{
					verts[v].texcoord[0] = DotProduct (vec, surf->texinfo->vecs[0]) * 0.015625f;
					verts[v].texcoord[1] = DotProduct (vec, surf->texinfo->vecs[1]) * 0.015625f;
				}
				else if (!(surf->flags & SURF_DRAWSKY))
				{
					verts[v].texcoord[0] = (DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]) / surf->texinfo->texture->width;
					verts[v].texcoord[1] = (DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]) / surf->texinfo->texture->height;
				}
			}

			glUnmapBuffer (GL_ARRAY_BUFFER);
		}
	}

	// for the next map
	glBindBuffer (GL_ARRAY_BUFFER, 0);
	r_numsurfaceverts = 0;
}


void R_RegisterSurface (msurface_t *surf)
{
	surf->numindexes = (surf->numvertexes - 2) * 3;
	surf->firstvertex = r_numsurfaceverts;
	r_numsurfaceverts += surf->numvertexes;
}

